# coding: utf-8
import json
import logging

from ws4py.server.handler.tornadohandler import WebSocketHandler

from sharedvariables import agents_list
from dbinterface import agent_login, save_log, set_installation_status, \
                        get_installation_log
from servertoagent import auth_request, auth_response, exit_request

browsers = []

def notify_browsers(msg='update', msg_type='out'):
    for ws in browsers:
        ws.write_message(json.dumps({'msg': msg, 'type': msg_type}))


class AgentsHandler(WebSocketHandler):
    def open(self):
        self._agent_name = ""
        self._logined = False
        self._no = 0
        self._state = 'Unknown'
        self._recent_notify = None
    
    def reset_counter(self):
        self._no = 0
        
    @property
    def status(self):
        return self._state
    
    def to_exit(self):
        self.write_message(exit_request())
        
    def on_close(self):             
        agents_list.pop(self._agent_name)
        self._set_state('Offline')
        
    def on_message(self, message):
        message = json.loads(message)
        command = message['command']
            
        if not self._logined and command != 'auth_response':
            self.write_message(auth_request())
        
        if command == 'auth_response':
            self._auth(message['login'], message['password'])
        elif command == 'state_response':
            self._set_state(message['state'])
        elif command == 'change_installation_state':
            if message['ended'] == 'Yes':
                set_installation_status(self._agent_name, message['recipe_name'], message['recipe_version'], message['status'], 
                                        get_installation_log(self._agent_name, message['recipe_name'], message['recipe_version']))
            else:
                set_installation_status(self._agent_name, message['recipe_name'], message['recipe_version'], message['status'])
        elif command == 'log_message':
            self._save_log(message['recipe_name'], message['recipe_version'], message['msg_type'], message['msg'])
                
    def _auth(self, login, password):
        self._logined = True
        
        if self._logined:
            agents_list[login] = self
            
            self._agent_name = login
            agent_login(login)
            self.write_message(auth_response(True))
        else:
            self.write_message(auth_response(False))
            self.write_message(auth_request())

    def _set_state(self, state):
        self._state = state
        if state != 'Installation':
            notify_browsers()
        print 'Agent', self._agent_name, 'has new state: ', state

    def _save_log(self, recipe_name, recipe_version, msg_type, msg):
        self._no += 1
        save_log(self._agent_name, recipe_name, recipe_version, msg_type, msg, self._no)
        notify_browsers(msg, msg_type)


class BrowsersHandler(WebSocketHandler):
    def open(self):
        browsers.append(self)

    def on_close(self):
        if self in browsers:
            browsers.remove(self)
