# coding: utf-8

import datetime

import pymongo

DB_NAME = 'testdb'
db = pymongo.Connection()[DB_NAME]


def agent_login(agent_id):
    agent = db.agents.find_one({'_id': agent_id})
    if not agent:
        name = agent_id
    else:
        name = agent['name']    
    db.agents.save({'_id': agent_id, 'name': name})
    
def get_agent_name(agent_id):
    agent = db.agents.find_one({'_id': agent_id})
    if agent:
        return agent['name']
    else:
        return None

def get_agents_names():
    agents = db.agents.find()
    return [a['name'] for a in agents]
    
def rename_agent(agent_id, new_name):
    agent = db.agents.find_one({'_id': agent_id})
    agent['name'] = new_name
    db.agents.save(agent) 

def delete_agent(agent_id):
    db.agents.remove({'_id': agent_id})
    
def get_agents_list():
    res = []
    for a in db.agents.find():
        res.append({'id': a['_id'], 'name': a['name']})
    return res

def new_installation(server_name, recipe_name, recipe_version, config, hidden_config_keys):
    db.log.remove({'server_name': server_name, 'recipe_name': recipe_name, 'recipe_version': recipe_version})
    db_config = {}
    db_config.update(config)
    for h in hidden_config_keys:
        del db_config[h]
    db.installed.save({'server_name': server_name, 
                       'recipe_name': recipe_name,
                       'recipe_version': recipe_version,
                       'date': str(datetime.datetime.now())[:19],
                       'status': 'sent',
                       'config': db_config})
        
def get_recipes_on_server(server_id):
    recipes = []
    for r in db.installed.find({'server_name': server_id}):
        if r['recipe_name'] not in recipes:
            recipes.append(r['recipe_name'])
    return recipes

def get_installations_by_recipe(server_id, recipe_name):
    installations = db.installed.find({'server_name': server_id, 'recipe_name': recipe_name})
    return [dict(i) for i in installations]

def get_installations(server_id):
    recipes_on_server = get_recipes_on_server(server_id)
    installations = []
    for recipe_name in recipes_on_server:
        installations.append((recipe_name, get_installations_by_recipe(server_id, recipe_name)))
    return installations
    
def save_log(server_name, recipe_name, recipe_version, msg_type, msg, n):
    db.log.save({'server_name': server_name, 
                 'recipe_name': recipe_name,
                 'recipe_version': recipe_version,
                 'type': msg_type,
                 'message': msg,
                 'n': n})    
    
def get_installation_log(server_name, recipe_name, recipe_version):
    res = db.log.find({'server_name': server_name, 'recipe_name': recipe_name, 'recipe_version': recipe_version})
    log = []
    for r in res:
        log.append((r['n'], r['type'], r['message']))
        
    def cmp_func(a, b):
        if int(a[0]) > int(b[0]):
            return 1
        elif int(b[0]) == int(a[0]):
            return 0
        else:
            return -1
            
    log.sort(cmp=cmp_func)
    return log

def set_installation_status(server_name, recipe_name, recipe_version, status, log=None):
    res = db.installed.find({'server_name': server_name, 'recipe_name': recipe_name, 'recipe_version': recipe_version})
    for r in res:
        to_change = r
    to_change['status'] = status
    if log is not None:
        to_change['log'] = log
    db.installed.save(to_change)
