# -*- coding: utf-8 -*-

from datetime import date
import json
import glob
import os
import time
from settings import RECIPES_PATH
from settings import BASE_PATH
from xml.etree import ElementTree

def get_version_number(version_file_name):
    return version_file_name[1:-4]

def get_default_recipe_info(version_file_name):
    return {"file": version_file_name, "who": "", "when": ""}

def get_recipe_names():
    recipes = [s for s in glob.glob(os.path.join(RECIPES_PATH, '*')) if os.path.isdir(s)]
    recipes.sort()
    return recipes

def get_recipes():
    res = []
    recipes = [s for s in glob.glob(os.path.join(RECIPES_PATH, '*')) if os.path.isdir(s)]
    recipes.sort()
    for r in recipes:
        versions_files = [v.split('/')[-1] for v in glob.glob(os.path.join(r, 'v*.xml')) if os.path.isfile(v)]
        try:
            versions_desc = json.load(open(os.path.join(RECIPES_PATH, r, 'about.json'), 'r'))
            versions_names_in_json = []
            for v in versions_desc:
                versions_names_in_json.append(v['file'])
            versions = []
            for v in versions_files:
                if v in versions_names_in_json:
                    for i in range(0, len(versions_desc)):
                        if versions_desc[i]['file'] == v:
                            versions.append(versions_desc[i])
                            break
                else:
                    versions.append(get_default_recipe_info(v))
        except Exception:
            versions = [get_default_recipe_info(v) for v in versions_files]
        
        for v in versions:
            v['name'] = get_version_number(v['file'])
        versions.sort(reverse=True)
        res.append((r.split('/')[-1], versions))
    return res

def get_next_version(recipe_name, recipes):
    for name, versions in recipes:
        if name == recipe_name:
            max_version = 0
            for v in versions:
                if int(v['name']) > max_version:
                    max_version = int(v['name'])
            return str(max_version + 1)
    return '1'
    

def get_default_config(recipe_name, recipe_version):
    tree = ElementTree.parse(get_recipe_file_name(recipe_name, recipe_version))
    conf = []
    for arg in tree.findall('./Args/Arg'):
        name = arg.text
        default = arg.get('default', '')
        if arg.get('type') == 'password':
            conf.append((name, default, 'password'))
        else:
            conf.append((name, default, 'text'))

    return conf

def get_config_list(recipe_name, config_dict):
    tree = ElementTree.parse(os.path.join(RECIPES_PATH, recipe_name))
    conf = []
    for arg in tree.findall('./Args/Arg'):
        name = arg.text
        if name in config_dict:
            conf.append(config_dict[name])
        else:
            conf.append('')
            
def not_valid_error_message(recipe):
    try:
        ElementTree.fromstring(recipe)
    except Exception as e:
        return e
    return None

def get_recipe_file_name(recipe_name, recipe_version):
    return os.path.join(RECIPES_PATH, recipe_name, 'v' + recipe_version + '.xml')

def create_recipe_path_if_not_exists(recipe_name):
    if not os.path.exists(os.path.join(RECIPES_PATH, recipe_name)):
        os.mkdir(os.path.join(RECIPES_PATH, recipe_name))
        
def get_date():
    d = date.today()
    return str(d.day) + '.' + str(d.month) + '.' + str(d.year) 
        
def recipe_save(name, version, content, username):
    file_ = open(get_recipe_file_name(name, version), 'w')
    file_.write(content.strip())
    file_.close()
    
    try:
        versions = json.load(open(os.path.join(RECIPES_PATH, name, 'about.json'), 'r'))
        version_names = [get_version_number(v['file']) for v in versions]
        if version in version_names:
            return 'Такая версия уже существует'
    except Exception as e:
        versions = []
        
    versions.append({'file': 'v' + str(version) + '.xml',
                   'who': username,
                   'when': get_date()})
    
    json.dump(versions, open(os.path.join(RECIPES_PATH, name, 'about.json'), 'w'))
    return None #ok
    
def get_recipe_content(name, version):
    file_ = open(get_recipe_file_name(name, version), 'r')
    content = file_.read()
    file_.close()
    return content

def get_recipe_template():
    file_ = open(os.path.join(RECIPES_PATH, "recipe_template.xml"), 'r')
    content = file_.read()
    file_.close()
    return content

def get_setup_tools():
    file_ = open(os.path.join(BASE_PATH, 'setup_tools.py'), 'r')  
    content = file_.read()
    file_.close()
    return content