# coding: utf-8

import os
from threading import Thread
import time
import logging

import tornado.ioloop
import tornado.web
import tornado.httpserver

from agentshandler import AgentsHandler, BrowsersHandler
from sharedvariables import agents_list
import settings

SOCKETS_PORT = 8001
STATIC_PATH = os.path.join(os.path.dirname(__file__), 'static')


class HttpServerThread(Thread):
    def __init__(self, urls, port, *args, **kwargs):
        Thread.__init__(self, *args, **kwargs)
        app = tornado.web.Application(urls)
        http_server = tornado.httpserver.HTTPServer(app)
        http_server.listen(port)
        self._stop = False
        self.daemon = True

    def stop(self):
        print 'Stop.'
        self._stop = self.__stop()
        self._loop.stop()
    
    def run(self):
        while not self._stop:
            try:
                self._loop = tornado.ioloop.IOLoop.instance()
                self._loop.start()
            except KeyboardInterrupt:
                print "Thread KeyboardInterrupt"
                break
    
def ensure_dir(d):
    if not os.path.exists(d):
        os.makedirs(d)
    
def start_server ():
    ensure_dir(settings.RECIPES_PATH)
    agents_urls = [(r'/agent-server', AgentsHandler),
                   (r'/socket', BrowsersHandler)]
    agents_server_thread = HttpServerThread(agents_urls, SOCKETS_PORT)
    agents_server_thread.start()
        
if __name__ == '__main__':
    start_server()
